import asyncio
import json
import traceback
from rsocket import Payload, BaseRequestHandler
from elarian.customer import Customer
from google.protobuf.json_format import MessageToJson
from .generated.messaging_model_pb2 import MessagingChannel,\
    MessagingSessionEndReason,\
    MessageReaction,\
    MessageDeliveryStatus
from .generated.payment_model_pb2 import PaymentChannel, PaymentStatus
from .generated.app_socket_pb2 import \
    ServerToAppNotification,\
    ServerToAppNotificationReply
from .generated.simulator_socket_pb2 import (
    ServerToSimulatorNotification,
    ServerToSimulatorNotificationReply
)
from .generated.common_model_pb2 import (
    CustomerNumberProvider,
)
from .helpers import (
    fill_in_outgoing_message,
    get_enum_string,
    has_key
)


class _RequestHandler(BaseRequestHandler):
    _handlers = dict()
    _is_simulator = False
    _client = None

    def register_handler(self, event, handler):
        self._handlers[event] = handler

    def handle(self, event, *args):
        if event in self._handlers.keys():
            handler = self._handlers[event]
            if asyncio.iscoroutinefunction(handler):
                loop = asyncio.get_event_loop()
                loop.create_task(handler(*args))
            else:
                handler(*args)

    def get_handlers(self):
        return self._handlers

    @staticmethod
    async def _default_handler(notif, customer, app_data, callback):
        print("No handler for notification:", notif)
        callback(data_update=app_data, response=None)

    async def request_response(self, payload: Payload) -> asyncio.Future:
        future = asyncio.Future()
        try:
            data = ServerToSimulatorNotification() if self._is_simulator else ServerToAppNotification()
            data.ParseFromString(payload.data)

            if data.WhichOneof("entry") == 'customer':
                data = data.customer
            elif data.WhichOneof("entry") == 'purse':
                data = data.purse

            event = data.WhichOneof("entry")
            notif = getattr(data, event)

            customer = None

            incoming_app_data = data.app_data if data.HasField('app_data') else None
            app_data = data.app_data if data.HasField('app_data') else None
            if app_data is not None:
                app_data = json.loads(app_data.string_val) if app_data.HasField('string_val') else app_data.bytes_val

            data = json.loads(MessageToJson(message=data, preserving_proto_field_name=True))
            notif = json.loads(MessageToJson(message=notif, preserving_proto_field_name=True))

            customer_number = notif.get('customer_number', None)

            # TODO: Complete notification parsing to match sdk spec

            if event in [
                'messaging_session_started',
                'messaging_session_renewed',
                'messaging_session_ended',
                'messaging_consent_update',
                'sent_message_reaction'
            ]:
                notif['channel_number']['channel'] = get_enum_string(
                    MessagingChannel,
                    notif['channel_number']['channel'],
                    'MESSAGING_CHANNEL'
                )
                if has_key('reason', notif):
                    notif['reason'] = get_enum_string(
                        MessagingSessionEndReason,
                        notif['reason'],
                        'MESSAGING_SESSION_END_REASON'
                    )
                if has_key('reaction', notif):
                    notif['reaction'] = get_enum_string(
                        MessageReaction,
                        notif['reaction'],
                        'MESSAGE_REACTION'
                    )

            if event == 'message_status':
                notif['status'] = get_enum_string(MessageDeliveryStatus, notif['status'], 'MESSAGE_DELIVERY_STATUS')

            if event in ['payment_status', 'wallet_payment_status', 'received_payment']:
                notif['status'] = get_enum_string(PaymentStatus, notif['status'], 'PAYMENT_STATUS')

                if has_key('channel_number', notif):
                    notif['channel_number']['channel'] = get_enum_string(
                        PaymentChannel,
                        notif['channel_number']['channel'],
                        'PAYMENT_CHANNEL'
                    )

                if has_key('customer_number', notif):
                    notif['customer_number']['provider'] = get_enum_string(
                        CustomerNumberProvider,
                        customer_number['provider'],
                        'CUSTOMER_NUMBER_PROVIDER')

            if event == 'received_message':
                notif['channel_number']['channel'] = get_enum_string(
                    MessagingChannel,
                    notif['channel_number']['channel'],
                    'MESSAGING_CHANNEL'
                )
                notif['customer_number']['provider'] = get_enum_string(
                            CustomerNumberProvider,
                            customer_number['provider'],
                            'CUSTOMER_NUMBER_PROVIDER')
                channel = notif['channel_number']['channel'].lower()
                if channel == 'sms':
                    event = 'received_sms'
                    notif['text'] = notif['parts'][0]['text']
                    del notif['parts']
                if channel == 'voice':
                    event = 'voice_call'
                if channel == 'ussd':
                    event = 'ussd_session'
                    notif['input'] = notif['parts'][0]['ussd']
                    del notif['parts']
                if channel == 'fb_messenger':
                    event = 'received_fb_messenger'
                if channel == 'telegram':
                    event = 'received_telegram'
                if channel == 'whatsapp':
                    event = 'received_whatsapp'
                if channel == 'email':
                    event = 'received_email'

            if event in self._handlers.keys():
                handler = self._handlers[event]
            else:
                handler = self._default_handler

            if not self._is_simulator:
                customer = Customer(client=self._client, id=data['customer_id'])
                if customer_number is not None:
                    customer = Customer(
                        client=self._client,
                        id=data['customer_id'],
                        number=customer_number['number'],
                        provider=customer_number['provider']
                    )
                # TODO: Fetch customer number from state if not available?

                notif['org_id'] = data['org_id']
                notif['app_id'] = data['app_id']
                notif['customer_id'] = customer.customer_id
                notif['created_at'] = data['created_at']

            def callback(response=None, data_update=None):
                res = ServerToSimulatorNotificationReply() if self._is_simulator else ServerToAppNotificationReply()
                if not self._is_simulator:
                    if response is not None:
                        res.message.CopyFrom(fill_in_outgoing_message(response))
                    if data_update is not None:
                        try:
                            ro = data_update.decode()
                            res.data_update.data.bytes_val = data_update
                        except (UnicodeDecodeError, AttributeError):
                            res.data_update.data.string_val = json.dumps(data_update)
                            pass
                    else:
                        res.data_update.data.CopyFrom(incoming_app_data)
                if not future.done():
                    future.set_result(Payload(data=res.SerializeToString(), metadata=bytes()))

            async def callback_timeout():
                await asyncio.sleep(15)
                if not future.done():
                    res = ServerToSimulatorNotificationReply() if self._is_simulator else ServerToAppNotificationReply()
                    if not self._is_simulator and incoming_app_data is not None:
                        res.data_update.data.CopyFrom(incoming_app_data)
                    future.set_result(Payload(data=res.SerializeToString(), metadata=bytes()))

            future.get_loop().create_task(handler(notif, customer, app_data or dict(), callback))
            future.get_loop().create_task(callback_timeout())

        except Exception as ex:
            traceback.print_exc()
            future.set_exception(ex)

        return future




import json
from google.protobuf.json_format import MessageToJson

from .client import Client
from .utils.helpers import fill_in_outgoing_message, has_key
from .utils.generated.app_socket_pb2 import (
    AppToServerCommand,
    AppToServerCommandReply,
    GenerateAuthTokenCommand,
)
from .utils.generated.common_model_pb2 import (
    CustomerNumberProvider,
)
from .utils.generated.payment_model_pb2 import (
    PaymentChannel,
    PaymentStatus,
)
from .utils.helpers import (
    get_enum_value,
    get_enum_string
)


class Elarian(Client):
    """Elarian class that allows setting of handlers to enable someone to deal with various situations.

        :param org_id: The organization id
        :param api_key: The generated API key from the dashboard
        :param app_id: The app id generated from the dashboard
    """

    def __init__(self, org_id, api_key, app_id, options=Client._default_options):
        super().__init__(
            org_id,
            api_key,
            app_id,
            [
                "reminder",
                "messaging_session_started",
                "messaging_session_renewed",
                "messaging_session_ended",
                "messaging_consent_update",
                "received_sms",
                "received_fb_messenger",
                "received_telegram",
                "received_whatsapp",
                "received_email",
                "voice_call",
                "ussd_session",
                "message_status",
                "sent_message_reaction",
                "received_payment",
                "payment_status",
                "wallet_payment_status",
                "customer_activity",
            ],
            options,
        )

    def set_on_reminder(self, handler):
        """Set the reminder handler.

        :param handler: Dedicated handler function
        """
        return self._on("reminder", handler)

    def set_on_messaging_session_started(self, handler):
        """Set the handler for session started notifications.

        :param handler: Dedicated handler function
        """
        return self._on("messaging_session_started", handler)

    def set_on_messaging_session_renewed(self, handler):
        """Set the handler for session renewed notifications.

        :param handler: Dedicated handler function
        """
        return self._on("messaging_session_renewed", handler)

    def set_on_messaging_session_ended(self, handler):
        """Set the handler for session ended notifications.

        :param handler: Dedicated handler function
        """
        return self._on("messaging_session_ended", handler)

    def set_on_messaging_consent_update(self, handler):
        """Set the handler for consent update notifications.

        :param handler: Dedicated handler function
        """
        return self._on("messaging_consent_update", handler)

    def set_on_received_sms(self, handler):
        """Set the handler for sms notifications.

        :param handler: Dedicated handler function
        """
        return self._on("received_sms", handler)

    def set_on_received_fb_messenger(self, handler):
        """Set the handler for messenger(facebook) notifications.

        :param handler: Dedicated handler function
        """
        return self._on("received_fb_messenger", handler)

    def set_on_received_telegram(self, handler):
        """Set the handler for telegram notifications.

        :param handler: Dedicated handler function
        """
        return self._on("received_telegram", handler)

    def set_on_received_whatsapp(self, handler):
        """Set the handler for whatsapp notifications.

        :param handler: Dedicated handler function
        """
        return self._on("received_whatsapp", handler)

    def set_on_received_email(self, handler):
        """Set the handler for email notifications.

        :param handler: Dedicated handler function
        """
        return self._on("received_email", handler)

    def set_on_voice_call(self, handler):
        """Set the handler for voice call notifications.

        :param handler: Dedicated handler function
        """
        return self._on("voice_call", handler)

    def set_on_ussd_session(self, handler):
        """Set the handler for ussd session notifications.

        :param handler: Dedicated handler function
        """
        return self._on("ussd_session", handler)

    def set_on_message_status(self, handler):
        """Set the handler for message status notifications.

        :param handler: Dedicated handler function
        """
        return self._on("message_status", handler)

    def set_on_sent_message_reaction(self, handler):
        """Set the handler for message reaction notifications.

        :param handler: Dedicated handler function
        """
        return self._on("sent_message_reaction", handler)

    def set_on_received_payment(self, handler):
        """Set the handler for payment notifications.

        :param handler: Dedicated handler function
        """
        return self._on("received_payment", handler)

    def set_on_payment_status(self, handler):
        """Set the handler for payment status notifications.

        :param handler: Dedicated handler function
        """
        return self._on("payment_status", handler)

    def set_on_wallet_payment_status(self, handler):
        """Set the handler for wallet payments status notifications.

        :param handler: Dedicated handler function
        """
        return self._on("wallet_payment_status", handler)

    def set_on_customer_activity(self, handler):
        """Set the handler for customer activity notifications.

        :param handler: Dedicated handler function
        """
        return self._on("customer_activity", handler)

    async def generate_auth_token(self):
        """Generate an auth token to use in place of API keys"""
        req = AppToServerCommand()
        req.generate_auth_token.CopyFrom(GenerateAuthTokenCommand())
        data = await self._send_command(req)
        res = self._parse_reply(data).generate_auth_token
        return {"token": res.token, "lifetime": res.lifetime.seconds}

    async def add_customer_reminder_by_tag(self, tag: dict, reminder: dict):
        """Set a reminder to be triggered at the specified time for customers with a particular tag.

        :param tag: Dictionary of tags
        :param reminder: Dictionary of reminders
        """
        req = AppToServerCommand()
        req.add_customer_reminder_tag.tag.key = tag["key"]
        req.add_customer_reminder_tag.tag.value.value = tag["value"]
        req.add_customer_reminder_tag.reminder.key = reminder["key"]
        req.add_customer_reminder_tag.reminder.remind_at.seconds = round(
            reminder["remind_at"]
        )
        if has_key("interval", reminder):
            req.add_customer_reminder_tag.reminder.interval.seconds = round(
                reminder["interval"]
            )
        req.add_customer_reminder_tag.reminder.payload.value = reminder["payload"]
        data = await self._send_command(req)
        res = self._parse_reply(data, to_json=True)['tag_command']
        if not res['status']:
            raise RuntimeError(res['description'])
        return res

    async def cancel_customer_reminder_by_tag(self, tag: dict, key: str):
        """Cancel a previously set reminder using a tag and key.

        :param tag: Dictionary of tags
        :param reminder: Dictionary of reminders
        """
        req = AppToServerCommand()
        req.cancel_customer_reminder_tag.key = key
        req.cancel_customer_reminder_tag.tag.key = tag["key"]
        req.cancel_customer_reminder_tag.tag.value.value = tag["value"]
        data = await self._send_command(req)
        res = self._parse_reply(data, to_json=True)['tag_command']
        if not res['status']:
            raise RuntimeError(res['description'])
        return res

    async def send_message_by_tag(
        self, tag: dict, messaging_channel: dict, message: dict
    ):
        """Send a message to customers with a specific tag.

        :param messaging_channel: Dictionary with the channles to be used
        :param message: Dictionary with the messages to be sent
        """
        req = AppToServerCommand()
        req.send_message_tag.channel_number.number = messaging_channel["number"]
        req.send_message_tag.channel_number.channel = messaging_channel["channel"].value
        req.send_message_tag.tag.key = tag["key"]
        req.send_message_tag.tag.value.value = tag["value"]
        req.send_message_tag.message.CopyFrom(fill_in_outgoing_message(message))
        data = await self._send_command(req)
        res = self._parse_reply(data, to_json=True)['tag_command']
        if not res['status']:
            raise RuntimeError(res['description'])
        return res

    async def initiate_payment(
        self, debit_party: dict, credit_party: dict, value: dict
    ):
        """Initiate a payment transaction.

        :param debit_party: Dictionary containing the details of the customer the money is coming from
        :param credit_party: Dictionary containing the details of the customer the money is going to
        :param value: Dictionary containing the amount and currency
        """
        req = AppToServerCommand()

        if has_key("purse", debit_party):
            req.initiate_payment.debit_party.purse.purse_id = debit_party["purse"]["purse_id"]
        elif has_key("customer", debit_party):
            req.initiate_payment.debit_party.customer.customer_number.number = (
                debit_party["customer"]["customer_number"]["number"]
            )
            req.initiate_payment.debit_party.customer.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                debit_party["customer"]["customer_number"]["provider"],
                "CUSTOMER_NUMBER_PROVIDER",
            )
            if has_key("partition", debit_party["customer"]["customer_number"]):
                req.initiate_payment.debit_party.customer.customer_number.partition = (
                    debit_party["customer"]["customer_number"]["partition"]
                )
            req.initiate_payment.debit_party.customer.channel_number.number = (
                debit_party["customer"]["channel_number"]["number"]
            )
            req.initiate_payment.debit_party.customer.channel_number.channel = get_enum_value(
                PaymentChannel,
                debit_party["customer"]["channel_number"]["channel"],
                "PAYMENT_CHANNEL",
            )
        elif has_key("wallet", debit_party):
            req.initiate_payment.debit_party.wallet.wallet_id = debit_party["wallet"]["wallet_id"]
            req.initiate_payment.debit_party.wallet.customer_id = debit_party["wallet"]["customer_id"]
        elif has_key("channel", debit_party):
            req.initiate_payment.debit_party.channel.channel_number.number = (
                debit_party["channel"]["channel_number"]["number"]
            )
            req.initiate_payment.debit_party.channel.channel_number.channel = get_enum_value(
                PaymentChannel,
                debit_party["customer"]["channel_number"]["channel"],
                "PAYMENT_CHANNEL",
            )
            req.initiate_payment.debit_party.channel.channel_code = debit_party["channel"]["network_code"]
            req.initiate_payment.debit_party.channel.account.value = debit_party["channel"]["account"].value

        if has_key("purse", credit_party):
            req.initiate_payment.credit_party.purse.purse_id = credit_party["purse"][
                "purse_id"
            ]
        elif has_key("customer", credit_party):
            req.initiate_payment.credit_party.customer.customer_number.number = (
                credit_party["customer"]["customer_number"]["number"]
            )
            req.initiate_payment.credit_party.customer.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                credit_party["customer"]["customer_number"]["provider"],
                "CUSTOMER_NUMBER_PROVIDER",
            )
            if has_key("partition", credit_party["customer"]["customer_number"]):
                req.initiate_payment.credit_party.customer.customer_number.partition = (
                    credit_party["customer"]["customer_number"]["partition"]
                )
            req.initiate_payment.credit_party.customer.channel_number.number = (
                credit_party["customer"]["channel_number"]["number"]
            )
            req.initiate_payment.credit_party.customer.channel_number.channel = get_enum_value(
                PaymentChannel,
                credit_party["customer"]["channel_number"]["channel"],
                "PAYMENT_CHANNEL",
            )
        elif has_key("wallet", credit_party):
            req.initiate_payment.credit_party.wallet.wallet_id = credit_party["wallet"][
                "wallet_id"
            ]
            req.initiate_payment.credit_party.wallet.customer_id = credit_party["wallet"]["customer_id"]
        elif has_key("channel", credit_party):
            req.initiate_payment.credit_party.channel.channel_number.number = (
                credit_party["channel"]["channel_number"]["number"]
            )
            req.initiate_payment.credit_party.channel.channel_number.channel = get_enum_value(
                PaymentChannel,
                credit_party["customer"]["channel_number"]["channel"],
                "PAYMENT_CHANNEL",
            )
            req.initiate_payment.credit_party.channel.channel_code = credit_party[
                "channel"
            ]["network_code"]
            req.initiate_payment.credit_party.channel.account.value = credit_party[
                "channel"
            ]["account"].value

        req.initiate_payment.value.amount = value["amount"]
        req.initiate_payment.value.currency_code = value["currency_code"]
        data = await self._send_command(req)
        res = self._parse_reply(data, to_json=True)['initiate_payment']
        res['status'] = get_enum_string(PaymentStatus, res['status'], 'PAYMENT_STATUS'),
        return res

    @staticmethod
    def _parse_reply(payload, to_json=False):
        result = AppToServerCommandReply()
        result.ParseFromString(payload.data)
        if to_json:
            result = json.loads(MessageToJson(message=result, preserving_proto_field_name=True))
        return result



from .generated.common_model_pb2 import (
    CustomerNumber,
    MediaType,
)
from .generated.messaging_model_pb2 import (
    OutboundMessage,
    VoiceCallAction,
    RecordSessionCallAction,
    RejectCallAction,
    PromptMessageReplyAction,
    PromptMessageMenuItemBody,
    TextToSpeechVoice,
    MessagingChannel
)


def has_key(key: str, data: dict):
    return key in data.keys()


def fill_in_outgoing_message(message: dict):
    _message = OutboundMessage()
    _message.labels.extend(message.get('labels', []))
    _message.provider_tag.value = message.get('provider_tag', '')
    _message.reply_token.value = message.get('reply_token', '')
    if has_key('reply_prompt', message):
        _message.reply_prompt.action = get_enum_value(
            PromptMessageReplyAction,
            message.get('reply_prompt', {}).get('action', 'UNKNOWN'),
            'PROMPT_MESSAGE_REPLY_ACTION'
        )

        def make_prompt_item(item):
            re = PromptMessageMenuItemBody()
            if has_key('text', item):
                re.text = item['text']
            elif has_key('media', item):
                re.media.url = item['media']['url']
                re.media.media = get_enum_value(MediaType, item['media']['type'], 'MEDIA_TYPE')
            return re
        _message.reply_prompt.menu.extend(list(
            map(make_prompt_item, message.get('reply_prompt', {}).get('menu', []))
        ))

    body = message.get('body', {})

    if has_key('text', body):
        _message.body.text = body['text']

    if has_key('url', body):
        _message.body.url = body['url']

    if has_key('ussd', body):
        _message.body.ussd.text = body['ussd']['text']
        _message.body.ussd.is_terminal = body['ussd']['is_terminal']

    if has_key('media', body):
        _message.body.media.url = body['media']['url']
        _message.body.media.media = get_enum_value(MediaType, body['media']['type'], 'MEDIA_TYPE')

    if has_key('location', body):
        _message.body.location.latitude = body['location']['latitude']
        _message.body.location.longitude = body['location']['longitude']
        _message.body.location.label.value = body['location'].get('label')
        _message.body.location.address.value = body['location'].get('address')

    if has_key('template', body):
        _message.body.template.id = body['template']['id']
        for k in body['template']['params']:
            _message.body.template.params[k] = body['template']['params'][k]

    if has_key('email', body):
        _message.body.email.subject = body['email'].get('subject')
        _message.body.email.body_plain = body['email'].get('plain')
        _message.body.email.body_html = body['email'].get('html')
        _message.body.email.cc_list.extend(body['email'].get('cc', []))
        _message.body.email.bcc_list.extend(body['email'].get('bcc', []))
        _message.body.email.attachments.extend(body['email'].get('attachments', []))

    if has_key('voice', body):
        for action in body['voice']:
            _action = VoiceCallAction()

            if has_key('say', action):
                _action.say.text = action['say']['text']
                _action.say.play_beep = action['say'].get('play_beep', False)
                _action.say.voice = get_enum_value(
                    TextToSpeechVoice,
                    action['say'].get('voice', 'FEMALE'),
                    'TEXT_TO_SPEECH_VOICE'
                )

            if has_key('play', action):
                _action.play.url = action['play']['url']

            if has_key('get_digits', action):
                _action.get_digits.num_digits.value = action['get_digits'].get('num_digits')
                _action.get_digits.timeout.seconds = action['get_digits'].get('timeout', 0)
                _action.get_digits.finish_on_key.value = action['get_digits'].get('finish_on_key', '#')

                if has_key('say', action.get_digits):
                    _action.get_digits.say.text = action['get_digits']['say']['text']
                    _action.get_digits.say.play_beep = action['get_digits']['say'].get('play_beep', False)
                    _action.get_digits.say.voice = get_enum_value(
                        TextToSpeechVoice,
                        action['get_digits']['say'].get('voice', 'FEMALE'),
                        'TEXT_TO_SPEECH_VOICE'
                    )

                if has_key('play', action.get_digits):
                    _action.get_digits.play.url = action['get_digits']['play']['url']

            if has_key('get_recording', action):
                _action.get_recording.timeout.seconds = action['get_recording'].get('timeout', 0)
                _action.get_recording.max_length.seconds = action['get_recording'].get('max_length', 3600)
                _action.get_recording.finish_on_key.value = action['get_recording'].get('finish_on_key', '#')
                _action.get_recording.play_beep = action['get_recording'].get('play_beep', False)
                _action.get_recording.trim_silence = action['get_recording'].get('trim_silence', True)

                if has_key('say', action.get_recording):
                    _action.get_recording.say.text = action['get_recording']['say']['text']
                    _action.get_recording.say.play_beep = action['get_recording']['say'].get('play_beep', False)
                    _action.get_recording.say.voice = get_enum_value(
                        TextToSpeechVoice,
                        action['get_recording']['say'].get('voice', 'FEMALE'),
                        'TEXT_TO_SPEECH_VOICE'
                    )

                if has_key('play', action.get_recording):
                    _action.get_recording.play.url = action['get_recording']['play']['url']

            if has_key('dial', action):
                _action.dial.record = action['dial'].get('record', False)
                _action.dial.sequential = action['dial'].get('sequential', True)
                _action.dial.ringback_tone.value = action['dial'].get('ringback_tone', None)
                _action.dial.caller_id.value = action['dial']['caller_id']
                _action.dial.max_duration.value = action['dial'].get('max_duration', 3600)
                numbers = []
                for num in action['dial']['customer_numbers']:
                    _num = CustomerNumber()
                    _num.number = num['number']
                    _num.provider = num['provider'].value
                    if has_key('partition', num):
                        _num.partition.value = num['partition']
                    numbers.append(num)
                _action.dial.customer_numbers.extend(numbers)

            if has_key('record_session', action):
                _action.record_session.CopyFrom(RecordSessionCallAction())

            if has_key('enqueue', action):
                _action.enqueue.hold_music.value = action['enqueue']['hold_music']
                _action.enqueue.queue_name.value = action['enqueue']['queue_name']

            if has_key('dequeue', action):
                _action.dequeue.record = action['dequeue'].get('record', False)
                _action.dequeue.queue_name.value = action['dequeue']['queue_name']
                _action.dequeue.channel_number.number = action['dequeue']['channel']['number']
                _action.dequeue.channel_number.channel = get_enum_value(
                    MessagingChannel,
                    action['dequeue']['channel']['channel'],
                    'MESSAGING_CHANNEL'
                )

            if has_key('reject', action):
                _action.reject.CopyFrom(RejectCallAction())

            if has_key('redirect', action):
                _action.redirect.url = action['redirect']['url']

            _message.body.voice.actions.append(_action)

    return _message


def get_valid_keys(target, prefix):
    """Lists the valid keys to be used on specific proto"""
    keys = map(lambda key: key.replace(f'{prefix}_', ''), target.keys())
    return list(filter(lambda item: item != 'UNSPECIFIED', list(keys)))


def get_enum_value(target, key, prefix):
    """Used to get the provider given the enum"""
    try:
        return target.Value(f"{prefix}_{key.upper()}")
    except Exception:
        raise RuntimeError(f"Invalid key {key}")


def get_enum_string(target, value, prefix):
    """Used to get the provider value given the enum"""
    try:
        keys = target.keys()
        if value in keys:
            return value.replace(f"{prefix}_", "")
        raise ValueError
    except Exception:
        raise RuntimeError(f"Invalid value {value} for {target}")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: identity_state.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_model_pb2 as common__model__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='identity_state.proto',
  package='com.elarian.hera.proto',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14identity_state.proto\x12\x16\x63om.elarian.hera.proto\x1a\x12\x63ommon_model.proto\"\xa0\x02\n\rIdentityState\x12\x33\n\x04tags\x18\x01 \x03(\x0b\x32%.com.elarian.hera.proto.CustomerIndex\x12\x45\n\x08metadata\x18\x02 \x03(\x0b\x32\x33.com.elarian.hera.proto.IdentityState.MetadataEntry\x12<\n\rsecondary_ids\x18\x03 \x03(\x0b\x32%.com.elarian.hera.proto.CustomerIndex\x1aU\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x33\n\x05value\x18\x02 \x01(\x0b\x32$.com.elarian.hera.proto.DataMapValue:\x02\x38\x01\x62\x06proto3'
  ,
  dependencies=[common__model__pb2.DESCRIPTOR,])




_IDENTITYSTATE_METADATAENTRY = _descriptor.Descriptor(
  name='MetadataEntry',
  full_name='com.elarian.hera.proto.IdentityState.MetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='com.elarian.hera.proto.IdentityState.MetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.elarian.hera.proto.IdentityState.MetadataEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=272,
  serialized_end=357,
)

_IDENTITYSTATE = _descriptor.Descriptor(
  name='IdentityState',
  full_name='com.elarian.hera.proto.IdentityState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tags', full_name='com.elarian.hera.proto.IdentityState.tags', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='com.elarian.hera.proto.IdentityState.metadata', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='secondary_ids', full_name='com.elarian.hera.proto.IdentityState.secondary_ids', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_IDENTITYSTATE_METADATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=69,
  serialized_end=357,
)

_IDENTITYSTATE_METADATAENTRY.fields_by_name['value'].message_type = common__model__pb2._DATAMAPVALUE
_IDENTITYSTATE_METADATAENTRY.containing_type = _IDENTITYSTATE
_IDENTITYSTATE.fields_by_name['tags'].message_type = common__model__pb2._CUSTOMERINDEX
_IDENTITYSTATE.fields_by_name['metadata'].message_type = _IDENTITYSTATE_METADATAENTRY
_IDENTITYSTATE.fields_by_name['secondary_ids'].message_type = common__model__pb2._CUSTOMERINDEX
DESCRIPTOR.message_types_by_name['IdentityState'] = _IDENTITYSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

IdentityState = _reflection.GeneratedProtocolMessageType('IdentityState', (_message.Message,), {

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _IDENTITYSTATE_METADATAENTRY,
    '__module__' : 'identity_state_pb2'
    # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.IdentityState.MetadataEntry)
    })
  ,
  'DESCRIPTOR' : _IDENTITYSTATE,
  '__module__' : 'identity_state_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.IdentityState)
  })
_sym_db.RegisterMessage(IdentityState)
_sym_db.RegisterMessage(IdentityState.MetadataEntry)


_IDENTITYSTATE_METADATAENTRY._options = None
# @@protoc_insertion_point(module_scope)



import struct
from abc import ABCMeta
from enum import IntEnum

PROTOCOL_MAJOR_VERSION = 1
PROTOCOL_MINOR_VERSION = 0


class ErrorCode(IntEnum):
    INVALID_SETUP = 0x001,
    UNSUPPORTED_SETUP = 0x002,
    REJECTED_SETUP = 0x003,
    REJECTED_RESUME = 0x004,
    CONNECTION_ERROR = 0x101,
    CONNECTION_ERROR_NO_RETRY = 0x102,
    APPLICATION_ERROR = 0x201,
    REJECTED = 0x202,
    CANCELED = 0x203,
    INVALID = 0x204,
    RESERVED = 0xFFFFFFFF


class ParseError(ValueError):
    pass


class Type(IntEnum):
    SETUP = 1,
    LEASE = 2,
    KEEPALIVE = 3,
    REQUEST_RESPONSE = 4,
    REQUEST_FNF = 5,
    REQUEST_STREAM = 6,
    REQUEST_CHANNEL = 7,
    REQUEST_N = 8,
    CANCEL = 9,
    PAYLOAD = 10,
    ERROR = 11,
    METADATA_PUSH = 12,
    RESUME = 13,
    RESUME_OK = 14,
    EXT = 0xFFFF


class Frame(metaclass=ABCMeta):
    __slots__ = (
        'length', 'frame_type', 'flags', 'stream_id', 'metadata', 'data',
        'flags_ignore', 'flags_metadata',
        'flags_follows', 'flags_complete', 'metadata_only')

    _FLAG_IGNORE_BIT = 0x200
    _FLAG_METADATA_BIT = 0x100

    def __init__(self, frame_type):
        self.length = 0
        self.frame_type = frame_type
        self.flags = 0
        self.stream_id = 0
        self.metadata = b''
        self.data = b''

        self.flags_ignore = False
        self.flags_metadata = False
        self.flags_follows = False
        self.flags_complete = False

        self.metadata_only = False

    class Parser:

        def __init__(self, buffer, offset, limit):
            self.buffer = buffer
            self.offset = offset
            self.limit = limit

        def parse_header(self):
            available = self.limit - self.offset
            if available < 12:
                raise ParseError('Not enough bytes: {} vs 12'.format(available))
            length, frame_type, flags, stream_id = struct.unpack_from(
                '>IHHI', self.buffer, self.offset)
            self.offset += 12
            return length, frame_type, flags, stream_id

    # noinspection PyAttributeOutsideInit
    def parse_header(self, buffer, offset):
        self.length, = struct.unpack('>I', b'\x00' + buffer[offset:offset + 3])
        self.stream_id, self.frame_type, self.flags = struct.unpack_from(
            '>IBB', buffer, offset + 3)
        self.flags |= (self.frame_type & 3) << 8
        self.frame_type >>= 2
        self.flags_ignore = (self.flags & self._FLAG_IGNORE_BIT) != 0
        self.flags_metadata = (self.flags & self._FLAG_METADATA_BIT) != 0
        return 9

    # noinspection PyAttributeOutsideInit
    def parse_metadata(self, buffer, offset):
        if not self.flags_metadata:
            return 0
        if not self.metadata_only:
            length, = struct.unpack('>I', b'\x00' + buffer[offset:offset + 3])
            offset += 3
        else:
            length = self.length - offset + 3
        self.metadata = buffer[offset:offset+length]
        return length + (0 if self.metadata_only else 3)

    # noinspection PyAttributeOutsideInit
    def parse_data(self, buffer, offset):
        length = self.length - offset + 3
        self.data = buffer[offset:offset+length]
        return length

    @staticmethod
    def pack_string(buffer):
        return struct.pack('b', len(buffer)) + buffer

    def serialize(self, middle=b''):
        self.flags &= ~(self._FLAG_IGNORE_BIT | self._FLAG_METADATA_BIT)
        if self.flags_ignore:
            self.flags |= self._FLAG_IGNORE_BIT
        if self.metadata:
            self.flags_metadata = True
            self.flags |= self._FLAG_METADATA_BIT

        # compute the length of the frame: header + middle + suffix.
        self.length = 9 + len(middle)
        if self.flags_metadata and self.metadata:
            self.length += len(self.metadata)
            if not self.metadata_only:
                self.length += 3
        if not self.metadata_only:
            self.length += len(self.data)

        offset = 0
        buffer = bytearray(self.length)

        buffer[offset:offset+3] = struct.pack('>I', self.length - 3)[1:]
        offset += 3

        struct.pack_into('>I', buffer, offset, self.stream_id)
        offset += 4

        buffer[7] = (self.frame_type << 2) | (self.flags >> 8)
        buffer[8] = self.flags & 0xff
        offset += 2

        buffer[offset:offset+len(middle)] = middle[:]
        offset += len(middle)

        if self.flags_metadata and self.metadata:
            length = len(self.metadata)
            if not self.metadata_only:
                buffer[offset:offset+3] = struct.pack('>I', length)[1:]
                offset += 3
            buffer[offset:offset+length] = self.metadata[:]
            offset += length

        if not self.metadata_only:
            buffer[offset:offset+len(self.data)] = self.data[:]
            offset += len(self.data)

        return bytes(buffer)


class SetupFrame(Frame):
    __slots__ = (
        'major_version', 'minor_version',
        'keep_alive_milliseconds', 'max_lifetime_milliseconds',
        'token_length', 'resume_identification_token',
        'metadata_encoding', 'data_encoding',
        'flags_lease', 'flags_strict', 'flags_resume')

    _FLAG_LEASE_BIT = 0x40
    _FLAG_RESUME_BIT = 0x80

    def __init__(self):
        super().__init__(Type.SETUP)
        self.major_version = PROTOCOL_MAJOR_VERSION
        self.minor_version = PROTOCOL_MINOR_VERSION
        self.flags_lease = False
        self.flags_resume = False
        self.keep_alive_milliseconds = 30000
        self.max_lifetime_milliseconds = 120000

    # noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += self.parse_header(buffer, offset)
        self.flags_lease = (self.flags & self._FLAG_LEASE_BIT) != 0
        self.flags_resume = (self.flags & self._FLAG_RESUME_BIT) != 0

        (self.major_version, self.minor_version,
         self.keep_alive_milliseconds, self.max_lifetime_milliseconds) = (
            struct.unpack_from('>HHII', buffer, offset))
        offset += 12

        if self.flags_resume:
            self.token_length = struct.unpack_from('>H', buffer, offset)
            offset += 2
            self.resume_identification_token = (
                buffer[offset:offset+self.token_length])
            offset += self.token_length

        def unpack_string():
            nonlocal offset
            length, = struct.unpack_from('b', buffer, offset)
            result = buffer[offset + 1:offset + length + 1]
            offset += length + 1
            return result

        self.metadata_encoding = unpack_string()
        self.data_encoding = unpack_string()

        offset += self.parse_metadata(buffer, offset)
        offset += self.parse_data(buffer, offset)

    def serialize(self, middle=b''):
        self.flags &= ~(self._FLAG_LEASE_BIT | self._FLAG_RESUME_BIT)
        if self.flags_lease:
            self.flags |= self._FLAG_LEASE_BIT
        if self.flags_resume:
            self.flags |= self._FLAG_RESUME_BIT
        middle = struct.pack(
            '>HHII', self.major_version, self.minor_version,
            self.keep_alive_milliseconds, self.max_lifetime_milliseconds)
        if self.flags_resume:
            middle += struct.pack('>H', self.token_length)
            assert len(self.resume_identification_token) == self.token_length
            assert isinstance(self.resume_identification_token, bytes)
            middle += self.resume_identification_token
        middle += self.pack_string(self.metadata_encoding)
        middle += self.pack_string(self.data_encoding)
        return Frame.serialize(self, middle)


class ErrorFrame(Frame):
    __slots__ = 'error_code'

    def __init__(self):
        super().__init__(Type.ERROR)

# noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += self.parse_header(buffer, offset)
        self.error_code, = struct.unpack_from('>I', buffer, offset)
        offset += 4
        offset += self.parse_data(buffer, offset)

    def serialize(self, middle=b''):
        middle = struct.pack('>I', self.error_code)
        return Frame.serialize(self, middle)


class LeaseFrame(Frame):
    __slots__ = ('time_to_live', 'number_of_requests')

    def __init__(self):
        super().__init__(Type.LEASE)
        super().metadata_only = True

    # noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += self.parse_header(buffer, offset)
        self.time_to_live, self.number_of_requests = struct.unpack_from(
            '>II', buffer, offset)
        offset += self.parse_metadata(buffer, offset)

    def serialize(self, middle=b''):
        middle = struct.pack('>II', self.time_to_live, self.number_of_requests)
        return Frame.serialize(self, middle)


class KeepAliveFrame(Frame):
    __slots__ = ('flags_respond', "last_receive_position")

    _FLAG_RESPOND_BIT = 0x80

    def __init__(self):
        super().__init__(Type.KEEPALIVE)
        self.flags_respond = False
        self.last_receive_position = 0

    # noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += self.parse_header(buffer, offset)
        self.flags_respond = (self.flags & self._FLAG_RESPOND_BIT) != 0
        self.last_receive_position = struct.unpack_from('>Q', buffer, offset)[0]
        offset += 8
        offset += self.parse_data(buffer, offset)

    def serialize(self, middle=b''):
        self.flags &= ~self._FLAG_RESPOND_BIT
        if self.flags_respond:
            self.flags |= self._FLAG_RESPOND_BIT
        middle = struct.pack('>Q', self.last_receive_position)
        return Frame.serialize(self, middle)


class RequestFrame(Frame):
    __slots__ = 'flags_follows'

    _FLAG_FOLLOWS_BIT = 0x80

    def __init__(self, frame_type):
        super().__init__(frame_type)
        self.flags_follows = False

    # noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        length = self.parse_header(buffer, offset)
        self.flags_follows = (self.flags & self._FLAG_FOLLOWS_BIT) != 0
        return length

    def serialize(self, middle=b''):
        self.flags &= ~self._FLAG_FOLLOWS_BIT
        if self.flags_follows:
            self.flags |= self._FLAG_FOLLOWS_BIT
        return Frame.serialize(self, middle)


class RequestResponseFrame(RequestFrame):
    __slots__ = ()

    def __init__(self):
        super().__init__(Type.REQUEST_RESPONSE)

    # noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += RequestFrame.parse(self, buffer, offset)
        offset += self.parse_metadata(buffer, offset)
        offset += self.parse_data(buffer, offset)


class RequestFireAndForgetFrame(RequestFrame):
    __slots__ = ()

    def __init__(self):
        super().__init__(Type.REQUEST_FNF)

# noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += RequestFrame.parse(self, buffer, offset)
        offset += self.parse_metadata(buffer, offset)
        offset += self.parse_data(buffer, offset)


class RequestStreamFrame(RequestFrame):
    __slots__ = 'initial_request_n'

    def __init__(self):
        super().__init__(Type.REQUEST_STREAM)
        self.initial_request_n = 0

    # noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += RequestFrame.parse(self, buffer, offset)
        self.initial_request_n, = struct.unpack_from('>I', buffer, offset)
        offset += 4
        offset += self.parse_metadata(buffer, offset)
        offset += self.parse_data(buffer, offset)

    def serialize(self, middle=b''):
        middle = struct.pack('>I', self.initial_request_n)
        return RequestFrame.serialize(self, middle)


class RequestChannelFrame(RequestFrame):
    __slots__ = ('initial_request_n', 'flags_complete', 'flags_follows')

    _FLAG_COMPLETE_BIT = 0x40
    _FLAG_FOLLOWS_BIT = 0x80

    def __init__(self):
        super().__init__(Type.REQUEST_CHANNEL)
        self.flags_complete = False
        self.flags_initial = False

    # noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += RequestFrame.parse(self, buffer, offset)
        self.flags_complete = (self.flags & self._FLAG_COMPLETE_BIT) != 0
        self.flags_follows = (self.flags & self._FLAG_FOLLOWS_BIT) != 0
        self.initial_request_n, = struct.unpack_from('>I', buffer, offset)
        offset += 4
        offset += self.parse_metadata(buffer, offset)
        offset += self.parse_data(buffer, offset)

    def serialize(self, middle=b''):
        middle = struct.pack('>I', self.initial_request_n)
        return RequestFrame.serialize(self, middle)


class RequestNFrame(RequestFrame):
    __slots__ = 'request_n'

    def __init__(self):
        super().__init__(Type.REQUEST_N)

    # noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += self.parse_header(buffer, offset)
        self.request_n = struct.unpack_from('>I', buffer, offset)

    def serialize(self, middle=b''):
        middle = struct.pack('>I', self.request_n)
        return Frame.serialize(self, middle)


class CancelFrame(Frame):

    def __init__(self):
        super().__init__(Type.CANCEL)

    def parse(self, buffer, offset):
        self.parse_header(buffer, offset)


class PayloadFrame(Frame):
    __slots__ = ('flags_follows', 'flags_complete', 'flags_next')

    _FLAG_FOLLOWS_BIT = 0x80
    _FLAG_COMPLETE_BIT = 0x40
    _FLAG_NEXT_BIT = 0x20

    def __init__(self):
        super().__init__(Type.PAYLOAD)
        self.flags_follows = False
        self.flags_complete = False
        self.flags_next = False

    # noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += self.parse_header(buffer, offset)
        self.flags_follows = (self.flags & self._FLAG_FOLLOWS_BIT) != 0
        self.flags_complete = (self.flags & self._FLAG_COMPLETE_BIT) != 0
        self.flags_next = (self.flags & self._FLAG_NEXT_BIT) != 0
        offset += self.parse_metadata(buffer, offset)
        offset += self.parse_data(buffer, offset)

    def serialize(self, middle=b''):
        self.flags &= ~(self._FLAG_FOLLOWS_BIT | self._FLAG_COMPLETE_BIT |
                        self._FLAG_NEXT_BIT)
        if self.flags_follows:
            self.flags |= self._FLAG_FOLLOWS_BIT
        if self.flags_complete:
            self.flags |= self._FLAG_COMPLETE_BIT
        if self.flags_next or self.data or self.metadata:
            self.flags |= self._FLAG_NEXT_BIT
        return Frame.serialize(self)


class MetadataPushFrame(Frame):
    __slots__ = ()

    # noinspection PyDunderSlots
    def __init__(self):
        super().__init__(Type.METADATA_PUSH)
        self.metadata_only = True
        self.flags_metadata = True

    # noinspection PyAttributeOutsideInit
    def parse(self, buffer, offset):
        offset += self.parse_header(buffer, offset)
        offset += self.parse_metadata(buffer, offset)


_table = {
    Type.SETUP: SetupFrame,
    Type.LEASE: LeaseFrame,
    Type.KEEPALIVE: KeepAliveFrame,
    Type.REQUEST_RESPONSE: RequestResponseFrame,
    Type.REQUEST_FNF: RequestFireAndForgetFrame,
    Type.REQUEST_STREAM: RequestStreamFrame,
    Type.REQUEST_CHANNEL: RequestChannelFrame,
    Type.REQUEST_N: RequestNFrame,
    Type.CANCEL: CancelFrame,
    Type.PAYLOAD: PayloadFrame,
    Type.ERROR: ErrorFrame,
    Type.METADATA_PUSH: MetadataPushFrame,
    # Type.RESUME: ResumeFrame,
    # Type.RESUME_OK: ResumeOKFrame,
}


def parse(buffer, offset=0):
    # A full header is 3 (length) + 4 (stream) + 2 (type, flags) bytes.
    if len(buffer) < 9:
        raise ParseError('Frame too short: {} bytes'.format(len(buffer)))
    frame_type = struct.unpack_from('>B', buffer, offset + 7)[0] >> 2
    try:
        frame = _table[frame_type]()
        frame.parse(buffer, offset)
        return frame
    except KeyError:
        raise ParseError('Wrong frame type: {}'.format(frame_type))



from abc import ABCMeta, abstractmethod


class Publisher(metaclass=ABCMeta):
    @abstractmethod
    def subscribe(self, subscriber):
        pass


class DefaultPublisher(Publisher):
    def subscribe(self, subscriber):
        subscriber.on_error(RuntimeError("Not implemented"))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: activity_state.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from . import common_model_pb2 as common__model__pb2
from . import activity_model_pb2 as activity__model__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='activity_state.proto',
  package='com.elarian.hera.proto',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14\x61\x63tivity_state.proto\x12\x16\x63om.elarian.hera.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x12\x63ommon_model.proto\x1a\x14\x61\x63tivity_model.proto\"\xe0\x02\n\x14\x41\x63tivitySessionState\x12?\n\x0f\x63ustomer_number\x18\x01 \x01(\x0b\x32&.com.elarian.hera.proto.CustomerNumber\x12\x45\n\x0e\x63hannel_number\x18\x02 \x01(\x0b\x32-.com.elarian.hera.proto.ActivityChannelNumber\x12\x12\n\nsession_id\x18\x03 \x01(\t\x12\x0e\n\x06\x61pp_id\x18\x04 \x01(\t\x12<\n\nactivities\x18\x05 \x03(\x0b\x32(.com.elarian.hera.proto.CustomerActivity\x12.\n\ncreated_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nupdated_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x91\x01\n\rActivityState\x12>\n\x08sessions\x18\x01 \x03(\x0b\x32,.com.elarian.hera.proto.ActivitySessionState\x12@\n\x10\x63ustomer_numbers\x18\x02 \x03(\x0b\x32&.com.elarian.hera.proto.CustomerNumberb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,common__model__pb2.DESCRIPTOR,activity__model__pb2.DESCRIPTOR,])




_ACTIVITYSESSIONSTATE = _descriptor.Descriptor(
  name='ActivitySessionState',
  full_name='com.elarian.hera.proto.ActivitySessionState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='customer_number', full_name='com.elarian.hera.proto.ActivitySessionState.customer_number', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='channel_number', full_name='com.elarian.hera.proto.ActivitySessionState.channel_number', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_id', full_name='com.elarian.hera.proto.ActivitySessionState.session_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='app_id', full_name='com.elarian.hera.proto.ActivitySessionState.app_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='activities', full_name='com.elarian.hera.proto.ActivitySessionState.activities', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_at', full_name='com.elarian.hera.proto.ActivitySessionState.created_at', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='updated_at', full_name='com.elarian.hera.proto.ActivitySessionState.updated_at', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=124,
  serialized_end=476,
)


_ACTIVITYSTATE = _descriptor.Descriptor(
  name='ActivityState',
  full_name='com.elarian.hera.proto.ActivityState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sessions', full_name='com.elarian.hera.proto.ActivityState.sessions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='customer_numbers', full_name='com.elarian.hera.proto.ActivityState.customer_numbers', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=479,
  serialized_end=624,
)

_ACTIVITYSESSIONSTATE.fields_by_name['customer_number'].message_type = common__model__pb2._CUSTOMERNUMBER
_ACTIVITYSESSIONSTATE.fields_by_name['channel_number'].message_type = activity__model__pb2._ACTIVITYCHANNELNUMBER
_ACTIVITYSESSIONSTATE.fields_by_name['activities'].message_type = activity__model__pb2._CUSTOMERACTIVITY
_ACTIVITYSESSIONSTATE.fields_by_name['created_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ACTIVITYSESSIONSTATE.fields_by_name['updated_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ACTIVITYSTATE.fields_by_name['sessions'].message_type = _ACTIVITYSESSIONSTATE
_ACTIVITYSTATE.fields_by_name['customer_numbers'].message_type = common__model__pb2._CUSTOMERNUMBER
DESCRIPTOR.message_types_by_name['ActivitySessionState'] = _ACTIVITYSESSIONSTATE
DESCRIPTOR.message_types_by_name['ActivityState'] = _ACTIVITYSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ActivitySessionState = _reflection.GeneratedProtocolMessageType('ActivitySessionState', (_message.Message,), {
  'DESCRIPTOR' : _ACTIVITYSESSIONSTATE,
  '__module__' : 'activity_state_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.ActivitySessionState)
  })
_sym_db.RegisterMessage(ActivitySessionState)

ActivityState = _reflection.GeneratedProtocolMessageType('ActivityState', (_message.Message,), {
  'DESCRIPTOR' : _ACTIVITYSTATE,
  '__module__' : 'activity_state_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.ActivityState)
  })
_sym_db.RegisterMessage(ActivityState)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: payment_model.proto

from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from . import common_model_pb2 as common__model__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='payment_model.proto',
  package='com.elarian.hera.proto',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x13payment_model.proto\x12\x16\x63om.elarian.hera.proto\x1a\x1egoogle/protobuf/wrappers.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x12\x63ommon_model.proto\"_\n\x14PaymentChannelNumber\x12\x37\n\x07\x63hannel\x18\x01 \x01(\x0e\x32&.com.elarian.hera.proto.PaymentChannel\x12\x0e\n\x06number\x18\x02 \x01(\t\"\xa9\x01\n\x19PendingPaymentTransaction\x12.\n\ncreated_at\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12+\n\x05value\x18\x02 \x01(\x0b\x32\x1c.com.elarian.hera.proto.Cash\x12/\n\tconverted\x18\x03 \x01(\x0b\x32\x1c.com.elarian.hera.proto.Cash\"\xc4\x02\n\x0ePaymentBalance\x12\x15\n\rcurrency_code\x18\x01 \x01(\t\x12/\n\tavailable\x18\x02 \x01(\x0b\x32\x1c.com.elarian.hera.proto.Cash\x12,\n\x06\x61\x63tual\x18\x03 \x01(\x0b\x32\x1c.com.elarian.hera.proto.Cash\x12\x44\n\x07pending\x18\x04 \x03(\x0b\x32\x33.com.elarian.hera.proto.PaymentBalance.PendingEntry\x12\x13\n\x0bsequence_nr\x18\x05 \x01(\x04\x1a\x61\n\x0cPendingEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12@\n\x05value\x18\x02 \x01(\x0b\x32\x31.com.elarian.hera.proto.PendingPaymentTransaction:\x02\x38\x01\",\n\x18PaymentPurseCounterParty\x12\x10\n\x08purse_id\x18\x01 \x01(\t\"C\n\x19PaymentWalletCounterParty\x12\x13\n\x0b\x63ustomer_id\x18\x01 \x01(\t\x12\x11\n\twallet_id\x18\x02 \x01(\t\"\xa4\x01\n\x1bPaymentCustomerCounterParty\x12?\n\x0f\x63ustomer_number\x18\x01 \x01(\x0b\x32&.com.elarian.hera.proto.CustomerNumber\x12\x44\n\x0e\x63hannel_number\x18\x02 \x01(\x0b\x32,.com.elarian.hera.proto.PaymentChannelNumber\"\xa7\x01\n\x1aPaymentChannelCounterParty\x12\x44\n\x0e\x63hannel_number\x18\x01 \x01(\x0b\x32,.com.elarian.hera.proto.PaymentChannelNumber\x12\x14\n\x0c\x63hannel_code\x18\x02 \x01(\x05\x12-\n\x07\x61\x63\x63ount\x18\x03 \x01(\x0b\x32\x1c.google.protobuf.StringValue\"\xb6\x02\n\x13PaymentCounterParty\x12G\n\x08\x63ustomer\x18\x01 \x01(\x0b\x32\x33.com.elarian.hera.proto.PaymentCustomerCounterPartyH\x00\x12\x41\n\x05purse\x18\x02 \x01(\x0b\x32\x30.com.elarian.hera.proto.PaymentPurseCounterPartyH\x00\x12\x43\n\x06wallet\x18\x03 \x01(\x0b\x32\x31.com.elarian.hera.proto.PaymentWalletCounterPartyH\x00\x12\x45\n\x07\x63hannel\x18\x04 \x01(\x0b\x32\x32.com.elarian.hera.proto.PaymentChannelCounterPartyH\x00\x42\x07\n\x05party\"\xa3\x03\n\x12PaymentTransaction\x12\x16\n\x0etransaction_id\x18\x01 \x01(\t\x12,\n\x06\x61pp_id\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12@\n\x0b\x64\x65\x62it_party\x18\x04 \x01(\x0b\x32+.com.elarian.hera.proto.PaymentCounterParty\x12\x41\n\x0c\x63redit_party\x18\x05 \x01(\x0b\x32+.com.elarian.hera.proto.PaymentCounterParty\x12+\n\x05value\x18\x06 \x01(\x0b\x32\x1c.com.elarian.hera.proto.Cash\x12\x35\n\x06status\x18\x07 \x01(\x0e\x32%.com.elarian.hera.proto.PaymentStatus\x12.\n\ncreated_at\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nupdated_at\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp*O\n\x0ePaymentChannel\x12\x1f\n\x1bPAYMENT_CHANNEL_UNSPECIFIED\x10\x00\x12\x1c\n\x18PAYMENT_CHANNEL_CELLULAR\x10\x01*\xe1\x05\n\rPaymentStatus\x12\x1e\n\x1aPAYMENT_STATUS_UNSPECIFIED\x10\x00\x12\x19\n\x15PAYMENT_STATUS_QUEUED\x10\x64\x12\'\n#PAYMENT_STATUS_PENDING_CONFIRMATION\x10\x65\x12%\n!PAYMENT_STATUS_PENDING_VALIDATION\x10\x66\x12\x1c\n\x18PAYMENT_STATUS_VALIDATED\x10g\x12#\n\x1ePAYMENT_STATUS_INVALID_REQUEST\x10\xc8\x01\x12!\n\x1cPAYMENT_STATUS_NOT_SUPPORTED\x10\xc9\x01\x12&\n!PAYMENT_STATUS_INSUFFICIENT_FUNDS\x10\xca\x01\x12%\n PAYMENT_STATUS_APPLICATION_ERROR\x10\xcb\x01\x12\x1f\n\x1aPAYMENT_STATUS_NOT_ALLOWED\x10\xcc\x01\x12%\n PAYMENT_STATUS_DUPLICATE_REQUEST\x10\xcd\x01\x12!\n\x1cPAYMENT_STATUS_INVALID_PURSE\x10\xce\x01\x12\"\n\x1dPAYMENT_STATUS_INVALID_WALLET\x10\xcf\x01\x12.\n)PAYMENT_STATUS_DECOMMISSIONED_CUSTOMER_ID\x10\xab\x02\x12\x1b\n\x16PAYMENT_STATUS_SUCCESS\x10\xac\x02\x12 \n\x1bPAYMENT_STATUS_PASS_THROUGH\x10\xad\x02\x12\x1a\n\x15PAYMENT_STATUS_FAILED\x10\x90\x03\x12\x1d\n\x18PAYMENT_STATUS_THROTTLED\x10\x91\x03\x12\x1b\n\x16PAYMENT_STATUS_EXPIRED\x10\x92\x03\x12\x1c\n\x17PAYMENT_STATUS_REJECTED\x10\x93\x03\x12\x1c\n\x17PAYMENT_STATUS_REVERSED\x10\xf4\x03\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_wrappers__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,common__model__pb2.DESCRIPTOR,])

_PAYMENTCHANNEL = _descriptor.EnumDescriptor(
  name='PaymentChannel',
  full_name='com.elarian.hera.proto.PaymentChannel',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_CHANNEL_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_CHANNEL_CELLULAR', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1915,
  serialized_end=1994,
)
_sym_db.RegisterEnumDescriptor(_PAYMENTCHANNEL)

PaymentChannel = enum_type_wrapper.EnumTypeWrapper(_PAYMENTCHANNEL)
_PAYMENTSTATUS = _descriptor.EnumDescriptor(
  name='PaymentStatus',
  full_name='com.elarian.hera.proto.PaymentStatus',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_QUEUED', index=1, number=100,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_PENDING_CONFIRMATION', index=2, number=101,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_PENDING_VALIDATION', index=3, number=102,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_VALIDATED', index=4, number=103,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_INVALID_REQUEST', index=5, number=200,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_NOT_SUPPORTED', index=6, number=201,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_INSUFFICIENT_FUNDS', index=7, number=202,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_APPLICATION_ERROR', index=8, number=203,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_NOT_ALLOWED', index=9, number=204,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_DUPLICATE_REQUEST', index=10, number=205,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_INVALID_PURSE', index=11, number=206,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_INVALID_WALLET', index=12, number=207,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_DECOMMISSIONED_CUSTOMER_ID', index=13, number=299,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_SUCCESS', index=14, number=300,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_PASS_THROUGH', index=15, number=301,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_FAILED', index=16, number=400,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_THROTTLED', index=17, number=401,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_EXPIRED', index=18, number=402,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_REJECTED', index=19, number=403,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAYMENT_STATUS_REVERSED', index=20, number=500,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1997,
  serialized_end=2734,
)
_sym_db.RegisterEnumDescriptor(_PAYMENTSTATUS)

PaymentStatus = enum_type_wrapper.EnumTypeWrapper(_PAYMENTSTATUS)
PAYMENT_CHANNEL_UNSPECIFIED = 0
PAYMENT_CHANNEL_CELLULAR = 1
PAYMENT_STATUS_UNSPECIFIED = 0
PAYMENT_STATUS_QUEUED = 100
PAYMENT_STATUS_PENDING_CONFIRMATION = 101
PAYMENT_STATUS_PENDING_VALIDATION = 102
PAYMENT_STATUS_VALIDATED = 103
PAYMENT_STATUS_INVALID_REQUEST = 200
PAYMENT_STATUS_NOT_SUPPORTED = 201
PAYMENT_STATUS_INSUFFICIENT_FUNDS = 202
PAYMENT_STATUS_APPLICATION_ERROR = 203
PAYMENT_STATUS_NOT_ALLOWED = 204
PAYMENT_STATUS_DUPLICATE_REQUEST = 205
PAYMENT_STATUS_INVALID_PURSE = 206
PAYMENT_STATUS_INVALID_WALLET = 207
PAYMENT_STATUS_DECOMMISSIONED_CUSTOMER_ID = 299
PAYMENT_STATUS_SUCCESS = 300
PAYMENT_STATUS_PASS_THROUGH = 301
PAYMENT_STATUS_FAILED = 400
PAYMENT_STATUS_THROTTLED = 401
PAYMENT_STATUS_EXPIRED = 402
PAYMENT_STATUS_REJECTED = 403
PAYMENT_STATUS_REVERSED = 500



_PAYMENTCHANNELNUMBER = _descriptor.Descriptor(
  name='PaymentChannelNumber',
  full_name='com.elarian.hera.proto.PaymentChannelNumber',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='channel', full_name='com.elarian.hera.proto.PaymentChannelNumber.channel', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='number', full_name='com.elarian.hera.proto.PaymentChannelNumber.number', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=132,
  serialized_end=227,
)


_PENDINGPAYMENTTRANSACTION = _descriptor.Descriptor(
  name='PendingPaymentTransaction',
  full_name='com.elarian.hera.proto.PendingPaymentTransaction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='created_at', full_name='com.elarian.hera.proto.PendingPaymentTransaction.created_at', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.elarian.hera.proto.PendingPaymentTransaction.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='converted', full_name='com.elarian.hera.proto.PendingPaymentTransaction.converted', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=230,
  serialized_end=399,
)


_PAYMENTBALANCE_PENDINGENTRY = _descriptor.Descriptor(
  name='PendingEntry',
  full_name='com.elarian.hera.proto.PaymentBalance.PendingEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='com.elarian.hera.proto.PaymentBalance.PendingEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.elarian.hera.proto.PaymentBalance.PendingEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=629,
  serialized_end=726,
)

_PAYMENTBALANCE = _descriptor.Descriptor(
  name='PaymentBalance',
  full_name='com.elarian.hera.proto.PaymentBalance',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='currency_code', full_name='com.elarian.hera.proto.PaymentBalance.currency_code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='available', full_name='com.elarian.hera.proto.PaymentBalance.available', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actual', full_name='com.elarian.hera.proto.PaymentBalance.actual', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pending', full_name='com.elarian.hera.proto.PaymentBalance.pending', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sequence_nr', full_name='com.elarian.hera.proto.PaymentBalance.sequence_nr', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PAYMENTBALANCE_PENDINGENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=402,
  serialized_end=726,
)


_PAYMENTPURSECOUNTERPARTY = _descriptor.Descriptor(
  name='PaymentPurseCounterParty',
  full_name='com.elarian.hera.proto.PaymentPurseCounterParty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='purse_id', full_name='com.elarian.hera.proto.PaymentPurseCounterParty.purse_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=728,
  serialized_end=772,
)


_PAYMENTWALLETCOUNTERPARTY = _descriptor.Descriptor(
  name='PaymentWalletCounterParty',
  full_name='com.elarian.hera.proto.PaymentWalletCounterParty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='customer_id', full_name='com.elarian.hera.proto.PaymentWalletCounterParty.customer_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='wallet_id', full_name='com.elarian.hera.proto.PaymentWalletCounterParty.wallet_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=774,
  serialized_end=841,
)


_PAYMENTCUSTOMERCOUNTERPARTY = _descriptor.Descriptor(
  name='PaymentCustomerCounterParty',
  full_name='com.elarian.hera.proto.PaymentCustomerCounterParty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='customer_number', full_name='com.elarian.hera.proto.PaymentCustomerCounterParty.customer_number', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='channel_number', full_name='com.elarian.hera.proto.PaymentCustomerCounterParty.channel_number', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=844,
  serialized_end=1008,
)


_PAYMENTCHANNELCOUNTERPARTY = _descriptor.Descriptor(
  name='PaymentChannelCounterParty',
  full_name='com.elarian.hera.proto.PaymentChannelCounterParty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='channel_number', full_name='com.elarian.hera.proto.PaymentChannelCounterParty.channel_number', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='channel_code', full_name='com.elarian.hera.proto.PaymentChannelCounterParty.channel_code', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='account', full_name='com.elarian.hera.proto.PaymentChannelCounterParty.account', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1011,
  serialized_end=1178,
)


_PAYMENTCOUNTERPARTY = _descriptor.Descriptor(
  name='PaymentCounterParty',
  full_name='com.elarian.hera.proto.PaymentCounterParty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='customer', full_name='com.elarian.hera.proto.PaymentCounterParty.customer', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='purse', full_name='com.elarian.hera.proto.PaymentCounterParty.purse', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='wallet', full_name='com.elarian.hera.proto.PaymentCounterParty.wallet', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='channel', full_name='com.elarian.hera.proto.PaymentCounterParty.channel', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='party', full_name='com.elarian.hera.proto.PaymentCounterParty.party',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1181,
  serialized_end=1491,
)


_PAYMENTTRANSACTION = _descriptor.Descriptor(
  name='PaymentTransaction',
  full_name='com.elarian.hera.proto.PaymentTransaction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction_id', full_name='com.elarian.hera.proto.PaymentTransaction.transaction_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='app_id', full_name='com.elarian.hera.proto.PaymentTransaction.app_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='debit_party', full_name='com.elarian.hera.proto.PaymentTransaction.debit_party', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='credit_party', full_name='com.elarian.hera.proto.PaymentTransaction.credit_party', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.elarian.hera.proto.PaymentTransaction.value', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='com.elarian.hera.proto.PaymentTransaction.status', index=5,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_at', full_name='com.elarian.hera.proto.PaymentTransaction.created_at', index=6,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='updated_at', full_name='com.elarian.hera.proto.PaymentTransaction.updated_at', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1494,
  serialized_end=1913,
)

_PAYMENTCHANNELNUMBER.fields_by_name['channel'].enum_type = _PAYMENTCHANNEL
_PENDINGPAYMENTTRANSACTION.fields_by_name['created_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PENDINGPAYMENTTRANSACTION.fields_by_name['value'].message_type = common__model__pb2._CASH
_PENDINGPAYMENTTRANSACTION.fields_by_name['converted'].message_type = common__model__pb2._CASH
_PAYMENTBALANCE_PENDINGENTRY.fields_by_name['value'].message_type = _PENDINGPAYMENTTRANSACTION
_PAYMENTBALANCE_PENDINGENTRY.containing_type = _PAYMENTBALANCE
_PAYMENTBALANCE.fields_by_name['available'].message_type = common__model__pb2._CASH
_PAYMENTBALANCE.fields_by_name['actual'].message_type = common__model__pb2._CASH
_PAYMENTBALANCE.fields_by_name['pending'].message_type = _PAYMENTBALANCE_PENDINGENTRY
_PAYMENTCUSTOMERCOUNTERPARTY.fields_by_name['customer_number'].message_type = common__model__pb2._CUSTOMERNUMBER
_PAYMENTCUSTOMERCOUNTERPARTY.fields_by_name['channel_number'].message_type = _PAYMENTCHANNELNUMBER
_PAYMENTCHANNELCOUNTERPARTY.fields_by_name['channel_number'].message_type = _PAYMENTCHANNELNUMBER
_PAYMENTCHANNELCOUNTERPARTY.fields_by_name['account'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
_PAYMENTCOUNTERPARTY.fields_by_name['customer'].message_type = _PAYMENTCUSTOMERCOUNTERPARTY
_PAYMENTCOUNTERPARTY.fields_by_name['purse'].message_type = _PAYMENTPURSECOUNTERPARTY
_PAYMENTCOUNTERPARTY.fields_by_name['wallet'].message_type = _PAYMENTWALLETCOUNTERPARTY
_PAYMENTCOUNTERPARTY.fields_by_name['channel'].message_type = _PAYMENTCHANNELCOUNTERPARTY
_PAYMENTCOUNTERPARTY.oneofs_by_name['party'].fields.append(
  _PAYMENTCOUNTERPARTY.fields_by_name['customer'])
_PAYMENTCOUNTERPARTY.fields_by_name['customer'].containing_oneof = _PAYMENTCOUNTERPARTY.oneofs_by_name['party']
_PAYMENTCOUNTERPARTY.oneofs_by_name['party'].fields.append(
  _PAYMENTCOUNTERPARTY.fields_by_name['purse'])
_PAYMENTCOUNTERPARTY.fields_by_name['purse'].containing_oneof = _PAYMENTCOUNTERPARTY.oneofs_by_name['party']
_PAYMENTCOUNTERPARTY.oneofs_by_name['party'].fields.append(
  _PAYMENTCOUNTERPARTY.fields_by_name['wallet'])
_PAYMENTCOUNTERPARTY.fields_by_name['wallet'].containing_oneof = _PAYMENTCOUNTERPARTY.oneofs_by_name['party']
_PAYMENTCOUNTERPARTY.oneofs_by_name['party'].fields.append(
  _PAYMENTCOUNTERPARTY.fields_by_name['channel'])
_PAYMENTCOUNTERPARTY.fields_by_name['channel'].containing_oneof = _PAYMENTCOUNTERPARTY.oneofs_by_name['party']
_PAYMENTTRANSACTION.fields_by_name['app_id'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
_PAYMENTTRANSACTION.fields_by_name['debit_party'].message_type = _PAYMENTCOUNTERPARTY
_PAYMENTTRANSACTION.fields_by_name['credit_party'].message_type = _PAYMENTCOUNTERPARTY
_PAYMENTTRANSACTION.fields_by_name['value'].message_type = common__model__pb2._CASH
_PAYMENTTRANSACTION.fields_by_name['status'].enum_type = _PAYMENTSTATUS
_PAYMENTTRANSACTION.fields_by_name['created_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PAYMENTTRANSACTION.fields_by_name['updated_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
DESCRIPTOR.message_types_by_name['PaymentChannelNumber'] = _PAYMENTCHANNELNUMBER
DESCRIPTOR.message_types_by_name['PendingPaymentTransaction'] = _PENDINGPAYMENTTRANSACTION
DESCRIPTOR.message_types_by_name['PaymentBalance'] = _PAYMENTBALANCE
DESCRIPTOR.message_types_by_name['PaymentPurseCounterParty'] = _PAYMENTPURSECOUNTERPARTY
DESCRIPTOR.message_types_by_name['PaymentWalletCounterParty'] = _PAYMENTWALLETCOUNTERPARTY
DESCRIPTOR.message_types_by_name['PaymentCustomerCounterParty'] = _PAYMENTCUSTOMERCOUNTERPARTY
DESCRIPTOR.message_types_by_name['PaymentChannelCounterParty'] = _PAYMENTCHANNELCOUNTERPARTY
DESCRIPTOR.message_types_by_name['PaymentCounterParty'] = _PAYMENTCOUNTERPARTY
DESCRIPTOR.message_types_by_name['PaymentTransaction'] = _PAYMENTTRANSACTION
DESCRIPTOR.enum_types_by_name['PaymentChannel'] = _PAYMENTCHANNEL
DESCRIPTOR.enum_types_by_name['PaymentStatus'] = _PAYMENTSTATUS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PaymentChannelNumber = _reflection.GeneratedProtocolMessageType('PaymentChannelNumber', (_message.Message,), {
  'DESCRIPTOR' : _PAYMENTCHANNELNUMBER,
  '__module__' : 'payment_model_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.PaymentChannelNumber)
  })
_sym_db.RegisterMessage(PaymentChannelNumber)

PendingPaymentTransaction = _reflection.GeneratedProtocolMessageType('PendingPaymentTransaction', (_message.Message,), {
  'DESCRIPTOR' : _PENDINGPAYMENTTRANSACTION,
  '__module__' : 'payment_model_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.PendingPaymentTransaction)
  })
_sym_db.RegisterMessage(PendingPaymentTransaction)

PaymentBalance = _reflection.GeneratedProtocolMessageType('PaymentBalance', (_message.Message,), {

  'PendingEntry' : _reflection.GeneratedProtocolMessageType('PendingEntry', (_message.Message,), {
    'DESCRIPTOR' : _PAYMENTBALANCE_PENDINGENTRY,
    '__module__' : 'payment_model_pb2'
    # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.PaymentBalance.PendingEntry)
    })
  ,
  'DESCRIPTOR' : _PAYMENTBALANCE,
  '__module__' : 'payment_model_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.PaymentBalance)
  })
_sym_db.RegisterMessage(PaymentBalance)
_sym_db.RegisterMessage(PaymentBalance.PendingEntry)

PaymentPurseCounterParty = _reflection.GeneratedProtocolMessageType('PaymentPurseCounterParty', (_message.Message,), {
  'DESCRIPTOR' : _PAYMENTPURSECOUNTERPARTY,
  '__module__' : 'payment_model_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.PaymentPurseCounterParty)
  })
_sym_db.RegisterMessage(PaymentPurseCounterParty)

PaymentWalletCounterParty = _reflection.GeneratedProtocolMessageType('PaymentWalletCounterParty', (_message.Message,), {
  'DESCRIPTOR' : _PAYMENTWALLETCOUNTERPARTY,
  '__module__' : 'payment_model_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.PaymentWalletCounterParty)
  })
_sym_db.RegisterMessage(PaymentWalletCounterParty)

PaymentCustomerCounterParty = _reflection.GeneratedProtocolMessageType('PaymentCustomerCounterParty', (_message.Message,), {
  'DESCRIPTOR' : _PAYMENTCUSTOMERCOUNTERPARTY,
  '__module__' : 'payment_model_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.PaymentCustomerCounterParty)
  })
_sym_db.RegisterMessage(PaymentCustomerCounterParty)

PaymentChannelCounterParty = _reflection.GeneratedProtocolMessageType('PaymentChannelCounterParty', (_message.Message,), {
  'DESCRIPTOR' : _PAYMENTCHANNELCOUNTERPARTY,
  '__module__' : 'payment_model_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.PaymentChannelCounterParty)
  })
_sym_db.RegisterMessage(PaymentChannelCounterParty)

PaymentCounterParty = _reflection.GeneratedProtocolMessageType('PaymentCounterParty', (_message.Message,), {
  'DESCRIPTOR' : _PAYMENTCOUNTERPARTY,
  '__module__' : 'payment_model_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.PaymentCounterParty)
  })
_sym_db.RegisterMessage(PaymentCounterParty)

PaymentTransaction = _reflection.GeneratedProtocolMessageType('PaymentTransaction', (_message.Message,), {
  'DESCRIPTOR' : _PAYMENTTRANSACTION,
  '__module__' : 'payment_model_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.PaymentTransaction)
  })
_sym_db.RegisterMessage(PaymentTransaction)


_PAYMENTBALANCE_PENDINGENTRY._options = None
# @@protoc_insertion_point(module_scope)



import json
from google.protobuf.json_format import MessageToJson

from .utils.generated.app_socket_pb2 import (
    AppToServerCommand,
    AppToServerCommandReply,
)
from .utils.generated.common_model_pb2 import (
    CustomerIndex,
    IndexMapping,
    CustomerNumberProvider,
)
from .utils.generated.messaging_model_pb2 import (
    MessagingChannel,
    MessagingConsentUpdate,
    MessageDeliveryStatus
)
from .utils.generated.activity_model_pb2 import (
    ActivityChannel,
)
from .utils.helpers import has_key, fill_in_outgoing_message, get_enum_value, get_enum_string


class Customer:
    """Customer class that handles a single customer instance, must have one of customer_id or customer_number.

        :param customer_id: The Elarian generated customer id
        :param customer_number: The customer phone number
    """

    _customer_id: str = None
    _customer_number: dict = None

    _identity_state: dict = None
    _messaging_state: dict = None
    _payment_state: dict = None
    _activity_state: dict = None

    _client = None

    def __init__(self, client, id=None, number=None, provider=None):
        self._client = client
        if id is not None:
            self._customer_id = id
        if number is not None:
            provider = (
                provider if provider is not None else 'CELLULAR'
            )
            self._customer_number = {"number": number, "provider": provider}

        if self.customer_id is None and self.customer_number is None:
            raise RuntimeError("Either id or number is required")

    async def get_state(self):
        """Used to get the current customer state."""
        req = AppToServerCommand()

        if self.customer_number is not None and has_key("number", self.customer_number):
            req.get_customer_state.customer_number.number = self.customer_number.get("number")
            req.get_customer_state.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        elif self.customer_id is not None:
            req.get_customer_state.customer_id = self.customer_id

        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        data = await self._send_command(req)
        res = self._parse_reply(data)['get_customer_state']

        if not res['status']:
            raise RuntimeError(res['description'])
        res = res['data']
        return res

    async def adopt_state(self, other_customer):
        """Used to adopt another customer's state.

        :param other_customer: Dictionary containing the other customer details
        """
        req = AppToServerCommand()

        if self.customer_id is None:
            await self.get_state()

        req.adopt_customer_state.customer_id = self.customer_id

        if other_customer.get("customer_id") is not None:
            req.adopt_customer_state.customer_id = other_customer.get("customer_id")
        elif other_customer.get("number") is not None and has_key(
            "number", other_customer
        ):
            req.adopt_customer_state.other_customer_number.number = (
                other_customer.get("number")
            )
            req.adopt_customer_state.other_customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                other_customer.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_state']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def send_message(self, messaging_channel: dict, message: dict):
        """Used to send a message to this customer.

        :param messaging_channel: Dictionary containing the messaging channels
        :param message: Dictionary containg the message being sent
        """
        req = AppToServerCommand()
        req.send_message.channel_number.number = messaging_channel.get("number")
        req.send_message.channel_number.channel = get_enum_value(
            MessagingChannel,
            messaging_channel.get("channel", "UNSPECIFIED"),
            "MESSAGING_CHANNEL",
        )
        req.send_message.customer_number.number = self.customer_number.get("number")
        req.send_message.customer_number.provider = get_enum_value(
            CustomerNumberProvider,
            self.customer_number.get("provider", "CELLULAR"),
            "CUSTOMER_NUMBER_PROVIDER",
        )
        req.send_message.message.CopyFrom(fill_in_outgoing_message(message))
        data = await self._send_command(req)
        res = self._parse_reply(data)['send_message']
        res['status'] = get_enum_string(MessageDeliveryStatus, res['status'], 'MESSAGE_DELIVERY_STATUS')
        return res

    async def reply_to_message(self, message_id: str, message: dict):
        """Used to reply to a message from this customer.

        :param message_id: Specific message id being replied to
        :param message: Message being sent back
        """
        req = AppToServerCommand()
        req.reply_to_message.customer_id = self.customer_id
        req.reply_to_message.message_id = message_id
        req.reply_to_message.message.CopyFrom(fill_in_outgoing_message(message))
        data = await self._send_command(req)
        res = self._parse_reply(data)['send_message']
        res['status'] = get_enum_string(MessageDeliveryStatus, res['status'], 'MESSAGE_DELIVERY_STATUS')
        return res

    async def update_activity(self, activity_channel: dict, activity: dict):
        """Used to update a customer's activity.

        :param activity_channel: Dictionary containg the activity channel(s)
        :param activity: Dictionary containing the activities
        """
        req = AppToServerCommand()

        req.customer_activity.channel_number.number = activity_channel.get("number")
        req.customer_activity.channel_number.channel = get_enum_value(
            ActivityChannel,
            activity_channel.get("channel", "CELLULAR"),
            "ACTIVITY_CHANNEL",
        )
        req.customer_activity.customer_number.number = self.customer_number.get(
            "number"
        )
        req.customer_activity.customer_number.provider = get_enum_value(
            CustomerNumberProvider,
            self.customer_number.get("provider", "CELLULAR"),
            "CUSTOMER_NUMBER_PROVIDER",
        )

        req.customer_activity.key = activity.get("key")
        req.customer_activity.session_id = activity.get("session_id")
        req.customer_activity.properties['property'] = str(activity.get("properties"))

        data = await self._send_command(req)
        res = self._parse_reply(data)['customer_activity']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def update_messaging_consent(
        self,
        messaging_channel: dict,
        action: str = 'ALLOW',
    ):
        """Used to update a customer's engagement consent on this channel.

        :param messaging_channel: Dictionary containing the messaging channels
        :param action: Choice of messaging constent. Default vallue: ALLOW
        """
        req = AppToServerCommand()

        req.update_messaging_consent.channel_number.number = messaging_channel.get(
            "number"
        )
        req.update_messaging_consent.channel_number.channel = get_enum_value(
            MessagingChannel,
            messaging_channel.get("channel", "UNSPECIFIED"),
            "MESSAGING_CHANNEL",
        )
        req.update_messaging_consent.customer_number.number = self.customer_number.get(
            "number"
        )
        req.update_messaging_consent.customer_number.provider = get_enum_value(
            CustomerNumberProvider,
            self.customer_number.get("provider", "CELLULAR"),
            "CUSTOMER_NUMBER_PROVIDER",
        )
        req.update_messaging_consent.update = get_enum_value(MessagingConsentUpdate, action, 'MESSAGING_CONSENT_UPDATE')
        data = await self._send_command(req)
        res = self._parse_reply(data)['update_messaging_consent']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def lease_app_data(self):
        """Used to lease a customer's app data. """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.lease_customer_app_data.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.lease_customer_app_data.customer_number.number = (
                self.customer_number.get("number")
            )
            req.lease_customer_app_data.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        data = await self._send_command(req)
        res = self._parse_reply(data, to_json=False).lease_customer_app_data

        if not res.status:
            raise RuntimeError(res.description)

        if res.value.HasField('string_val'):
            return json.loads(res.string_val)
        else:
            return res.value.bytes_val

    async def update_app_data(self, data):
        """Used to update a customer's app data.

        :param data: Dictionary of the data being updated
        """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.update_customer_app_data.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.update_customer_app_data.customer_number.number = (
                self.customer_number.get("number")
            )
            req.update_customer_app_data.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        try:
            ro = data.decode()
            req.update_customer_app_data.update.bytes_val = data
        except (UnicodeDecodeError, AttributeError):
            req.update_customer_app_data.update.string_val = json.dumps(data)
            pass

        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_app_data']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def delete_app_data(self):
        """Used to remove a customer's app data. """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.delete_customer_app_data.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.delete_customer_app_data.customer_number.number = (
                self.customer_number.get("number")
            )
            req.delete_customer_app_data.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_app_data']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def get_metadata(self):
        """Used to get the metadata of the customer.
        """
        state = await self.get_state()
        meta = dict()
        if has_key('identity_state', state):
            meta = state['identity_state'].get('metadata', dict())
        result = dict()
        for key in meta:
            result[key] = json.loads(meta[key]['string_val']) if has_key('string_val', meta[key]) else meta[key]['bytes_val']
        return result

    async def update_metadata(self, data: dict):
        """Used to update a customer's metadata.

        :param data: Dictionary containing the metadata being updated
        """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.update_customer_metadata.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.update_customer_metadata.customer_number.number = (
                self.customer_number.get("number")
            )
            req.update_customer_metadata.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        for key in data.keys():
            val = data.get(key)
            try:
                data = val.decode()
                req.update_customer_metadata.updates[key].bytes_val = val
            except (UnicodeDecodeError, AttributeError):
                req.update_customer_metadata.updates[key].string_val = json.dumps(val)
                pass

        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_state']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def delete_metadata(self, keys: list):
        """Used to remove a customer's metadata.

        :param keys: List of keys being deleted
        """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.delete_customer_metadata.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.delete_customer_metadata.customer_number.number = (
                self.customer_number.get("number")
            )
            req.delete_customer_metadata.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        req.delete_customer_metadata.deletions.extend(keys)
        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_state']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def get_secondary_ids(self):
        """Used to get the secondary ids of the customer."""
        state = await self.get_state()
        if has_key('identity_state', state):
            return state['identity_state']['secondary_ids'] or list()
        return list()

    async def update_secondary_ids(self, secondary_ids: list):
        """Used to update a customer's secondary ids.

        :param secondary_id: List of secondary ids being updated
        """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.update_customer_secondary_id.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.update_customer_secondary_id.customer_number.number = (
                self.customer_number.get("number")
            )
            req.update_customer_secondary_id.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        for _id in secondary_ids:
            val = CustomerIndex()
            val.mapping.key = _id["key"]
            val.mapping.value.value = _id["value"]
            if _id["expires_at"] is not None:
                val.expires_at.seconds = _id["expires_at"]
            req.update_customer_secondary_id.updates.append(val)

        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_state']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def delete_secondary_ids(self, secondary_ids: list):
        """Used to remove a customer's secondary ids.

        :param secondary_ids: List of secondary ids being deleted
        """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.delete_customer_secondary_id.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.delete_customer_secondary_id.customer_number.number = (
                self.customer_number.get("number")
            )
            req.delete_customer_secondary_id.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        for _id in secondary_ids:
            val = IndexMapping()
            val.key = _id.get("key")
            val.value.value = _id.get("value")
            req.delete_customer_secondary_id.deletions.append(val)

        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_state']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def get_tags(self):
        """Used to get the tags for a customer."""
        state = await self.get_state()
        if has_key('identity_state', state):
            return state['identity_state']['tags'] or list()
        return list()

    async def update_tags(self, tags: list):
        """Used to update a customer's tags.

        :param tags: list of tags being updated
        """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.update_customer_tag.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.update_customer_tag.customer_number.number = self.customer_number.get(
                "number"
            )
            req.update_customer_tag.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        for _id in tags:
            val = CustomerIndex()
            val.mapping.key = _id.get("key")
            val.mapping.value.value = _id.get("value")
            if _id.get("expires_at") is not None:
                val.expires_at.seconds = _id.get("expires_at")
            req.update_customer_tag.updates.append(val)

        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_state']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def delete_tags(self, keys: list):
        """Used to remove a customer's tags.

        :param keys: list of tags being deleted
        """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.delete_customer_tag.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.delete_customer_tag.customer_number.number = self.customer_number.get(
                "number"
            )
            req.delete_customer_tag.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        req.delete_customer_tag.deletions.extend(keys)

        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_state']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def add_reminder(self, reminder: dict):
        """Used to add a reminder.

        :param reminder: dictionary containing the details of the reminder
        """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.add_customer_reminder.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.add_customer_reminder.customer_number.number = self.customer_number.get(
                "number"
            )
            req.add_customer_reminder.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        req.add_customer_reminder.reminder.key = reminder.get("key")
        req.add_customer_reminder.reminder.remind_at.seconds = round(
            reminder.get("remind_at")
        )
        req.add_customer_reminder.reminder.payload.value = reminder.get("payload")
        if reminder.get("interval") is not None:
            req.add_customer_reminder.reminder.interval.seconds = round(
                reminder.get("interval")
            )

        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_state']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def cancel_reminder(self, key: str):
        """Used to cancel a reminder based on keys.

        :param key: Reminder key
        """
        req = AppToServerCommand()

        if self.customer_id is not None:
            req.cancel_customer_reminder.customer_id = self.customer_id
        elif self.customer_number is not None and has_key(
            "number", self.customer_number
        ):
            req.cancel_customer_reminder.customer_number.number = (
                self.customer_number.get("number")
            )
            req.cancel_customer_reminder.customer_number.provider = get_enum_value(
                CustomerNumberProvider,
                self.customer_number.get("provider", "CELLULAR"),
                "CUSTOMER_NUMBER_PROVIDER",
            )
        else:
            raise RuntimeError("Invalid customer id and/or customer number")

        req.cancel_customer_reminder.key = key

        data = await self._send_command(req)
        res = self._parse_reply(data)['update_customer_app_data']

        if not res['status']:
            raise RuntimeError(res['description'])

        return res

    async def _send_command(self, data):
        return await self._client._send_command(data)

    @staticmethod
    def _parse_reply(payload, to_json=True):
        result = AppToServerCommandReply()
        result.ParseFromString(payload.data)
        if to_json:
            result = json.loads(MessageToJson(message=result, preserving_proto_field_name=True))
        return result




import ssl
import asyncio
import atexit
from abc import ABCMeta
from rsocket import RSocket, Payload
from .utils.generated.app_socket_pb2 import AppConnectionMetadata
from .utils.request_handler import _RequestHandler


class Client(metaclass=ABCMeta):
    """Client class that handles connections to the Elarian backend

        :param org_id: The organization id
        :param api_key: The generated API key from the dashboard
        :param app_id: The app id generated from the dashboard
    """
    _default_options = {
        "resumable": False,
        "lifetime": 60000,
        "keep_alive": 1000,
        "allow_notifications": True,
    }

    _org_id = None
    _app_id = None
    _api_key = None
    _is_simulator = False
    _options = _default_options

    _socket = None
    _connection = None
    _is_connected = False
    _loop = None
    _request_handler = _RequestHandler(None)

    def __init__(self, org_id: str, api_key: str, app_id: str, events: list, options: dict = None):
        self._app_id = app_id
        self._api_key = api_key
        self._org_id = org_id
        self._expected_events = events + ['pending', 'error', 'connecting', 'connected', 'closed']
        if options is not None:
            self._options.update(options)

    async def connect(self):
        """Used to connect to Elarian."""
        self._request_handler.handle("pending")

        setup = AppConnectionMetadata()
        setup.org_id = self._org_id
        setup.app_id = self._app_id
        setup.api_key.value = self._api_key
        setup.simplex_mode = not self._options['allow_notifications']
        setup.simulator_mode = self._is_simulator

        self._loop = asyncio.get_event_loop()
        self._connection = await asyncio.open_connection(
            host='tcp.elarian.dev',
            loop=self._loop,
            port=8082,
            ssl=ssl.create_default_context(),
            server_hostname='tcp.elarian.dev',
        )
        self._socket = RSocket(
            reader=self._connection[0],
            writer=self._connection[1],
            server=False,
            loop=self._loop,
            data_encoding=b'application/octet-stream',
            metadata_encoding=b'application/octet-stream',
            setup_payload=Payload(data=setup.SerializeToString()),
            keep_alive_milliseconds=self._options['keep_alive'],
            max_lifetime_milliseconds=self._options['lifetime'],
            handler_factory=self._make_request_handler,
            error_handler=self._error_handler
        )

        self._is_connected = True
        self._loop.create_task(self.__keep_running())

        self._request_handler.handle("connected")
        atexit.register(self.__clean_up)

        return self

    async def disconnect(self):
        """Used to disconnect from Elarian."""
        self._is_connected = False
        self._connection[1].close()
        await self._socket.close()
        self._request_handler.handle("closed")

    def is_connected(self):
        """Checks whether the connection has been successfully established. 

           :returns: True or False
        """
        return self._is_connected

    def set_on_connection_pending(self, handler):
        """Sets handler for when the connection is pending

           :param handler: Dedicated handler function
        """
        return self._on('pending', handler)

    def set_on_connection_error(self, handler):
        """Sets handler for when there is a connection error

            :param handler: Dedicated handler function
        """
        return self._on('error', handler)

    def set_on_connection_closed(self, handler):
        """Sets handler for when the connection is closed

            :param handler: Dedicated handler function
        """
        return self._on('closed', handler)

    def set_on_connecting(self, handler):
        """Sets handler for when the connection is being opened

            :param handler: Dedicated handler function
        """
        return self._on('connecting', handler)

    def set_on_connected(self, handler):
        """Sets handler for when the connection has been successfully established

            :param handler: Dedicated handler function
        """
        return self._on('connected', handler)

    def _on(self, event, handler):
        if event in self._expected_events:
            self._request_handler.register_handler(event, handler)
        else:
            raise RuntimeError("Unexpected event {0}. Must be one of {1}".format(event, self._expected_events))
        return self

    def _make_request_handler(self, socket):
        registered = self._request_handler.get_handlers()
        self._request_handler = _RequestHandler(socket)
        self._request_handler._is_simulator = self._is_simulator
        self._request_handler._client = self
        for event in registered.keys():
            self._request_handler.register_handler(event, registered[event])
        self._request_handler.handle("connecting")
        return self._request_handler

    def _error_handler(self, error: Payload):
        self._request_handler.handle("error", RuntimeError(error.data.decode()))
        self._loop.create_task(self.disconnect())

    def __clean_up(self):
        if self.is_connected():
            if self._loop.is_running():
                self._loop.create_task(self.disconnect())
            else:
                asyncio.run(self.disconnect())

    async def __keep_running(self):
        while self.is_connected():
            await asyncio.sleep(2)

    async def _send_command(self, data):
        if not self.is_connected():
            raise RuntimeError("Client is not connected")
        return await self._socket.request_response(self._make_payload(data))

    @staticmethod
    def _make_payload(data):
        return Payload(data=data.SerializeToString(), metadata=bytes())



import json
from google.protobuf.json_format import MessageToJson

from .client import Client

from .utils.generated.simulator_socket_pb2 import (
    SimulatorToServerCommand,
    SimulatorToServerCommandReply,
)
from .utils.generated.messaging_model_pb2 import (
    InboundMessageBody,
    MessagingChannel,
)

from .utils.generated.common_model_pb2 import (
    MediaType
)

from .utils.generated.payment_model_pb2 import (
    PaymentChannel,
    PaymentStatus
)
from .utils.helpers import has_key, get_enum_value


class Simulator(Client):
    """Client class that simulates a connection to the elarian backend

        :param org_id: The organization id
        :param api_key: The generated API key from the dashboard
        :param app_id: The app id generated from the dashboard
    """

    def __init__(self, org_id, api_key, app_id, options=Client._default_options):
        super().__init__(
            org_id,
            api_key,
            app_id,
            [
                "send_message",
                "make_voice_call",
                "send_customer_payment",
                "send_channel_payment",
                "checkout_payment",
            ],
            options,
        )
        self._is_simulator = True

    def set_on_send_message(self, handler):
        """Set the handler for messages received on the simulator

           :param handler: Dedicated handler function
        """
        return self._on("send_message", handler)

    def set_on_make_voice_call(self, handler):
        """Set the handler for voice calls received on the simulator

           :param handler: Dedicated handler function
        """
        return self._on("make_voice_call", handler)

    def set_on_send_customer_payment(self, handler):
        """Set the handler for customer payments received on the simulator

           :param handler: Dedicated handler function
        """
        return self._on("send_customer_payment", handler)

    def set_on_send_channel_payment(self, handler):
        """Set the handler for channel payments received on the simulator

           :param handler: Dedicated handler function
        """
        return self._on("send_channel_payment", handler)

    def set_on_checkout_payment(self, handler):
        """Set the handler for customer checkout received on the simulator

           :param handler: Dedicated handler function
        """
        return self._on("checkout_payment", handler)

    async def receive_message(
        self,
        phone_number: str,
        messaging_channel: dict,
        session_id: str,
        message_parts: list,
        cost: dict = None
    ):
        """Simulate sending a message. i.e. tell the simulated gateway to receive a message

           :param phone_number: Phone number that sent the message
           :param messaging_channel: Dictionary containing the details of the messaging channel used
           :param session_id: Session id of the simulation
           :param message_parts: List containing the message parts
        """
        req = SimulatorToServerCommand()

        if cost is None:
            cost = {
                "currency_code": "KES",
                "amount": 0
            }
        req.receive_message.session_id.value = session_id
        req.receive_message.customer_number = phone_number
        req.receive_message.channel_number.number = messaging_channel["number"]
        req.receive_message.cost.amount = cost["amount"]
        req.receive_message.cost.currency_code = cost["currency_code"]
        req.receive_message.channel_number.channel = get_enum_value(
            MessagingChannel,
            messaging_channel["channel"],
            "MESSAGING_CHANNEL",
        )

        parts = req.receive_message.parts
        for part in message_parts:
            _part = InboundMessageBody()

            if has_key("text", part):
                _part.text = part["text"]

            if has_key("ussd", part):
                _part.ussd.value = part["ussd"]

            if has_key("media", part):
                _part.media.url = part["media"]["url"]
                _part.media.media = get_enum_value(
                    MediaType,
                    part["media"]["type"],
                    "MEDIA_TYPE",
                )

            if has_key("location", part):
                _part.location.latitude = part["location"]["latitude"]
                _part.location.longitude = part["location"]["longitude"]
                _part.location.label.value = part["location"]["label"]
                _part.location.address.value = part["location"]["address"]

            if has_key("email", part):
                _part.email.subject = part["email"]["subject"]
                _part.email.body_plain = part["email"]["plain"]
                _part.email.body_html = part["email"]["html"]
                _part.email.cc_list.extend(part["email"]["cc"])
                _part.email.bcc_list.extend(part["email"]["bcc"])
                _part.email.attachments.extend(part["email"]["attachments"])

            if has_key("voice", part):
                _part.voice.direction = part["voice"]["direction"].value
                _part.voice.status = part["voice"]["status"].value
                _part.voice.started_at.seconds = part["voice"]["started_at"]
                _part.voice.hangup_cause = part["voice"]["hangup_cause"].value
                _part.voice.dtmf_digits.value = part["voice"]["dtmf_digits"]
                _part.voice.recording_url.value = part["voice"]["recording_url"]

                _part.voice.dial_data.destination_number = part["voice"]["dial_data"][
                    "destination_number"
                ]
                _part.voice.dial_data.started_at.seconds = part["voice"]["dial_data"][
                    "started_at"
                ]
                _part.voice.dial_data.duration.seconds = part["voice"]["dial_data"][
                    "duration"
                ]

                _part.voice.queue_data.enqueued_at.seconds = part["voice"][
                    "queue_data"
                ]["enqueued_at"]
                _part.voice.queue_data.dequeued_at.seconds = part["voice"][
                    "queue_data"
                ]["dequeued_at"]
                _part.voice.queue_data.dequeued_to_number.value = part["voice"][
                    "queue_data"
                ]["dequeued_to_number"]
                _part.voice.queue_data.dequeued_to_sessionId.value = part["voice"][
                    "queue_data"
                ]["dequeued_to_sessionId"]
                _part.voice.queue_data.queue_duration.seconds = part["voice"][
                    "queue_data"
                ]["queue_duration"]

            parts.append(_part)
        data = await self._send_command(req)
        res = self._parse_reply(data)
        return res

    async def receive_payment(
        self,
        phone_number: str,
        payment_channel: dict,
        transaction_id: str,
        value: dict,
        status: str,
    ):
        """Used to simulate the receiving of a payment

           :param phone_number: Phone number that sent the payment
           :param payment_channel: Dictionary containing the details of the payment channel used
           :param transaction_id: Transaction id of the simulation
           :param value: Dictionary containing the transaction value and currency
           :param status: Status of the transaction
           """
        req = SimulatorToServerCommand()
        req.receive_payment.transaction_id = transaction_id
        req.receive_payment.customer_number = phone_number
        req.receive_payment.status = get_enum_value(
            PaymentStatus,
            status,
            "PAYMENT_STATUS",
        )
        req.receive_payment.value.amount = value["amount"]
        req.receive_payment.value.currency_code = value["currency_code"]
        req.receive_payment.channel_number.number = payment_channel["number"]
        req.receive_payment.channel_number.channel = get_enum_value(
            PaymentChannel,
            payment_channel["channel"],
            "PAYMENT_CHANNEL",
        )
        data = await self._send_command(req)
        res = self._parse_reply(data)
        return res

    async def update_payment_status(self, transaction_id: str, status: str):
        """Used to simulate the updating of a payment status

           :param transaction_id: Transaction id of the simulation
           :param status: Status of the transaction
        """
        req = SimulatorToServerCommand()
        req.update_payment_status.transaction_id = transaction_id
        req.update_payment_status.status = get_enum_value(
            PaymentStatus,
            status,
            "PAYMENT_STATUS",
        )
        data = await self._send_command(req)
        res = self._parse_reply(data)
        return res

    @staticmethod
    def _parse_reply(payload, to_json=True):
        result = SimulatorToServerCommandReply()
        result.ParseFromString(payload.data)
        if to_json:
            result = json.loads(MessageToJson(message=result, preserving_proto_field_name=True))
        return result



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: app_state.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import common_model_pb2 as common__model__pb2
import app_model_pb2 as app__model__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='app_state.proto',
  package='com.elarian.hera.proto',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0f\x61pp_state.proto\x12\x16\x63om.elarian.hera.proto\x1a\x12\x63ommon_model.proto\x1a\x0f\x61pp_model.proto\"{\n\x08\x41ppState\x12\x32\n\x04\x64\x61ta\x18\x01 \x01(\x0b\x32$.com.elarian.hera.proto.DataMapValue\x12;\n\treminders\x18\x02 \x03(\x0b\x32(.com.elarian.hera.proto.CustomerReminderb\x06proto3'
  ,
  dependencies=[common__model__pb2.DESCRIPTOR,app__model__pb2.DESCRIPTOR,])




_APPSTATE = _descriptor.Descriptor(
  name='AppState',
  full_name='com.elarian.hera.proto.AppState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='com.elarian.hera.proto.AppState.data', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reminders', full_name='com.elarian.hera.proto.AppState.reminders', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=80,
  serialized_end=203,
)

_APPSTATE.fields_by_name['data'].message_type = common__model__pb2._DATAMAPVALUE
_APPSTATE.fields_by_name['reminders'].message_type = app__model__pb2._CUSTOMERREMINDER
DESCRIPTOR.message_types_by_name['AppState'] = _APPSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AppState = _reflection.GeneratedProtocolMessageType('AppState', (_message.Message,), {
  'DESCRIPTOR' : _APPSTATE,
  '__module__' : 'app_state_pb2'
  # @@protoc_insertion_point(class_scope:com.elarian.hera.proto.AppState)
  })
_sym_db.RegisterMessage(AppState)


# @@protoc_insertion_point(module_scope)



import struct

from rsocket import frame

__all__ = ['Connection']


class Connection:
    def __init__(self):
        self._buffer = bytearray()

    def receive_data(self, data: bytes):
        self._buffer.extend(data)
        total = len(self._buffer)
        events = []

        while total >= 3:
            length, = struct.unpack('>I', b'\x00' + self._buffer[:3])
            if total < length + 3:
                break
            events.append(frame.parse(self._buffer[:length + 3]))
            self._buffer = self._buffer[length + 3:]
            total -= length + 3
        return events



import asyncio


class Timer:
    def __init__(self, timeout, callback, callback_args=()):
        self._timeout = timeout
        self._callback = callback
        self._task = asyncio.ensure_future(self._job(callback_args))

    async def _job(self, callback_args):
        await asyncio.sleep(self._timeout)
        await self._callback(*callback_args)

    def cancel(self):
        self._task.cancel()


class IntervalTimer:
    def __init__(self, interval, callback, callback_args=()):
        self._interval = interval
        self._callback = callback
        self._ok = True
        self._task = asyncio.ensure_future(self._job(callback_args))

    async def _job(self, callback_args):
        try:
            while self._ok:
                await asyncio.sleep(self._interval)
                await self._callback(*callback_args)
        except Exception as ex:
            print(ex)

    def cancel(self):
        self._ok = False
        self._task.cancel()


from abc import ABCMeta, abstractmethod


class Subscriber(metaclass=ABCMeta):
    @abstractmethod
    def on_subscribe(self, subscription):
        pass

    @abstractmethod
    def on_next(self, value):
        pass

    @abstractmethod
    def on_error(self, exception):
        pass

    @abstractmethod
    def on_complete(self):
        pass


class DefaultSubscriber(Subscriber):
    def on_next(self, value):
        pass

    def on_error(self, exception):
        pass

    def on_subscribe(self, subscription):
        pass

    def on_complete(self):
        pass